using System;
using System.Collections.Generic;

using Syntriax.Engine.Core;
using Syntriax.Engine.Core.Abstract;
using Syntriax.Engine.Physics2D.Abstract;

namespace Syntriax.Engine.Physics2D;

public class PhysicsEngine2D : IPhysicsEngine2D
{
    private readonly List<IRigidBody2D> rigidBodies = new(32);
    private readonly List<ICollider2D> colliders = new(64);

    private int _iterationCount = 1;
    private ICollisionDetector collisionDetector = new CollisionDetector();

    public int IterationCount { get => _iterationCount; set => _iterationCount = value < 1 ? 1 : value; }

    public void AddRigidBody(IRigidBody2D rigidBody)
    {
        if (rigidBodies.Contains(rigidBody))
            return;

        rigidBodies.Add(rigidBody);

        foreach (var collider2D in rigidBody.BehaviourController.GetBehaviours<ICollider2D>())
            colliders.Add(collider2D);

        rigidBody.BehaviourController.OnBehaviourAdded += OnBehaviourAdded;
        rigidBody.BehaviourController.OnBehaviourRemoved += OnBehaviourRemoved;
    }

    public void RemoveRigidBody(IRigidBody2D rigidBody)
    {
        rigidBodies.Remove(rigidBody);
    }

    public void Step(float deltaTime)
    {
        float intervalDeltaTime = deltaTime / IterationCount;

        for (int iterationIndex = 0; iterationIndex < IterationCount; iterationIndex++)
        {
            // Can Parallel
            for (int i = 0; i < rigidBodies.Count; i++)
                StepRigidBody(rigidBodies[i], intervalDeltaTime);

            // Can Parallel
            foreach (var collider in colliders)
                collider.Recalculate();

            // Can Parallel
            for (int x = 0; x < colliders.Count; x++)
            {
                ICollider2D? colliderX = colliders[x];
                for (int y = 0; y < colliders.Count; y++)
                {
                    ICollider2D? colliderY = colliders[y];

                    if (colliderX.RigidBody2D == colliderY.RigidBody2D && colliderY.RigidBody2D is null)
                        continue;

                    if (collisionDetector.TryDetect(colliderX, colliderY, out CollisionDetectionInformation? information))
                    {
                        information.Left.Transform.Position -= .5f * information.Normal * information.Penetration;
                        information.Right.Transform.Position += .5f * information.Normal * information.Penetration;
                    }
                }
            }
        }
    }

    private static void StepRigidBody(IRigidBody2D rigidBody, float intervalDeltaTime)
    {
        rigidBody.Transform.Position += rigidBody.Velocity * intervalDeltaTime;
        rigidBody.Transform.Rotation += rigidBody.AngularVelocity * intervalDeltaTime;
    }

    private static void CacheRotations(List<IRigidBody2D> rigidBodies, List<float> rotations)
    {
        rotations.Clear();
        for (int i = 0; i < rigidBodies.Count; i++)
            rotations.Add(rigidBodies[i].Transform.Rotation);
    }

    private static void CachePositions(List<IRigidBody2D> rigidBodies, List<Vector2D> positions)
    {
        positions.Clear();
        for (int i = 0; i < rigidBodies.Count; i++)
            positions.Add(rigidBodies[i].Transform.Position);
    }

    private void OnBehaviourAdded(IBehaviourController controller, IBehaviour behaviour)
    {
        if (behaviour is not ICollider2D collider2D)
            return;

        colliders.Add(collider2D);
    }

    private void OnBehaviourRemoved(IBehaviourController controller, IBehaviour behaviour)
    {
        if (behaviour is not ICollider2D collider2D)
            return;

        colliders.Remove(collider2D);
    }
}
